@@ -35,6 +35,7 @@
 
 #include <test/support/tdb_catch.h>
 #include "test/support/src/helpers.h"
+#include "test/support/src/vfs_helpers.h"
 #include "tiledb/sm/c_api/tiledb_serialization.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
 #include "tiledb/sm/cpp_api/tiledb"
@@ -55,23 +56,8 @@
 using namespace tiledb;
 using ResultSetType = std::map<std::string, std::any>;
 
-using tiledb::test::ServerQueryBuffers;
-using tiledb::test::submit_query_wrapper;
-
 namespace {
 
-#ifdef _WIN32
-static const char PATH_SEPARATOR = '\\';
-static std::string current_dir() {
-  return sm::Win::current_dir();
-}
-#else
-static const char PATH_SEPARATOR = '/';
-static std::string current_dir() {
-  return sm::Posix::current_dir();
-}
-#endif
-
 template <class T>
 bool check_result(const T a, const T b, size_t start, size_t end) {
   auto a_exp = T(a.begin() + start, a.begin() + end);
@@ -100,34 +86,15 @@ bool check_result(
 }
 
 struct SerializationFx {
-  const std::string tmpdir = "serialization_test_dir";
-  const std::string array_name = "testarray";
-  const std::string array_uri =
-      current_dir() + PATH_SEPARATOR + tmpdir + "/" + array_name;
-
+  test::VFSTestSetup vfs_test_setup_;
+  tiledb_ctx_t* ctx_;
   Context ctx;
-  VFS vfs;
-
-  // Serialization parameters
-  bool serialize_ = true;
-  bool refactored_query_v2_ = false;
-  bool finalize_ = false;
-  // Buffers to allocate on server side for serialized queries
-  tiledb::test::ServerQueryBuffers server_buffers_;
+  const std::string array_uri;
 
   SerializationFx()
-      : vfs(ctx) {
-    if (vfs.is_dir(tmpdir))
-      vfs.remove_dir(tmpdir);
-    vfs.create_dir(tmpdir);
-    if (!vfs.is_dir(tmpdir))
-      std::cerr << "'created' but not finding dir '" << tmpdir << "'"
-                << std::endl;
-  }
-
-  ~SerializationFx() {
-    if (vfs.is_dir(tmpdir))
-      vfs.remove_dir(tmpdir);
+      : ctx_{vfs_test_setup_.ctx_c}
+      , ctx{vfs_test_setup_.ctx()}
+      , array_uri{vfs_test_setup_.array_uri("testarray")} {
   }
 
   static void check_read_stats(const Query& query) {
@@ -226,15 +193,7 @@ struct SerializationFx {
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
 
     // The deserialized query should also include the write stats
     check_write_stats(query);
@@ -278,20 +237,12 @@ struct SerializationFx {
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
   }
 
   void write_sparse_array() {
-    std::vector<int32_t> coords = {1, 1, 2, 2, 3, 3, 4, 4, 5,  5,
-                                   6, 6, 7, 7, 8, 8, 9, 9, 10, 10};
+    std::vector<int32_t> d1 = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
+    std::vector<int32_t> d2 = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
     std::vector<uint32_t> a1;
     std::vector<uint32_t> a2;
     std::vector<uint8_t> a2_nullable;
@@ -315,7 +266,8 @@ struct SerializationFx {
     Array array(ctx, array_uri, TILEDB_WRITE);
     Query query(ctx, array);
     query.set_layout(TILEDB_UNORDERED);
-    query.set_coordinates(coords);
+    query.set_data_buffer("d1", d1);
+    query.set_data_buffer("d2", d2);
     query.set_data_buffer("a1", a1);
     query.set_data_buffer("a2", a2);
     query.set_validity_buffer("a2", a2_nullable);
@@ -323,15 +275,7 @@ struct SerializationFx {
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
   }
 
   void write_sparse_array_split_coords() {
@@ -371,15 +315,7 @@ struct SerializationFx {
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
   }
 };
 
@@ -388,8 +324,12 @@ struct SerializationFx {
 TEST_CASE_METHOD(
     SerializationFx,
     "subarray - Query serialization, dense",
-    "[query][dense][serialization]") {
-  refactored_query_v2_ = GENERATE(true, false);
+    "[query][dense][serialization][rest][sc-40489]") {
+  if (!vfs_test_setup_.is_rest()) {
+    SUCCEED("sc-40489: this test passes on remote arrays only");
+    return;
+  }
+
   create_array(TILEDB_DENSE);
   auto expected_results = write_dense_array();
   check_subarray_stats(2, 2);
@@ -417,15 +357,7 @@ TEST_CASE_METHOD(
     check_subarray_stats(3, 3);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     // Check stats after serialization
@@ -465,15 +397,7 @@ TEST_CASE_METHOD(
     query.set_condition(condition);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     // The deserialized query should also include the write stats
@@ -535,15 +459,7 @@ TEST_CASE_METHOD(
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
     check_subarray_stats(5, 5);
 
@@ -578,15 +494,7 @@ TEST_CASE_METHOD(
 
     // Submit initial query.
     set_buffers(query);
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     check_subarray_stats(5, 5);
 
     REQUIRE(query.query_status() == Query::Status::INCOMPLETE);
@@ -599,15 +507,7 @@ TEST_CASE_METHOD(
 
     // Reset buffers, serialize and resubmit
     set_buffers(query);
-    rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     check_subarray_stats(7, 7);
 
     REQUIRE(query.query_status() == Query::Status::INCOMPLETE);
@@ -620,15 +520,7 @@ TEST_CASE_METHOD(
 
     // Reset buffers, serialize and resubmit
     set_buffers(query);
-    rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     check_subarray_stats(9, 9);
 
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
@@ -644,15 +536,13 @@ TEST_CASE_METHOD(
 TEST_CASE_METHOD(
     SerializationFx,
     "subarray - Query serialization, sparse",
-    "[query][sparse][serialization]") {
-  refactored_query_v2_ = GENERATE(true, false);
+    "[query][sparse][serialization][rest]") {
   create_array(TILEDB_SPARSE);
   write_sparse_array();
 
   SECTION("- Read all") {
     Array array(ctx, array_uri, TILEDB_READ);
     Query query(ctx, array);
-    std::vector<int32_t> coords(1000);
     std::vector<uint32_t> a1(1000);
     std::vector<uint32_t> a2(1000);
     std::vector<uint8_t> a2_nullable(1000);
@@ -664,7 +554,6 @@ TEST_CASE_METHOD(
     cppapi_subarray.set_subarray(subarray);
     query.set_subarray(cppapi_subarray);
 
-    query.set_coordinates(coords);
     query.set_data_buffer("a1", a1);
     query.set_data_buffer("a2", a2);
     query.set_validity_buffer("a2", a2_nullable);
@@ -672,15 +561,7 @@ TEST_CASE_METHOD(
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     auto result_el = query.result_buffer_elements_nullable();
@@ -696,7 +577,6 @@ TEST_CASE_METHOD(
     SerializationFx,
     "subarray - Query serialization, split coords, sparse",
     "[query][sparse][serialization][split-coords]") {
-  refactored_query_v2_ = GENERATE(true, false);
   create_array(TILEDB_SPARSE);
   write_sparse_array_split_coords();
 
@@ -722,15 +602,7 @@ TEST_CASE_METHOD(
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     auto result_el = query.result_buffer_elements_nullable();
@@ -746,8 +618,7 @@ TEST_CASE_METHOD(
 TEST_CASE_METHOD(
     SerializationFx,
     "subarray - Query serialization, dense ranges",
-    "[query][dense][serialization]") {
-  refactored_query_v2_ = GENERATE(true, false);
+    "[query][dense][serialization][rest]") {
   create_array(TILEDB_DENSE);
   write_dense_array_ranges();
 
@@ -772,15 +643,7 @@ TEST_CASE_METHOD(
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     auto result_el = query.result_buffer_elements_nullable();
@@ -812,15 +675,7 @@ TEST_CASE_METHOD(
     query.set_offsets_buffer("a3", a3_offsets);
 
     // Submit query
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
 
     auto result_el = query.result_buffer_elements_nullable();
@@ -855,15 +710,7 @@ TEST_CASE_METHOD(
 
     // Submit initial query.
     set_buffers(query);
-    auto rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
     REQUIRE(query.query_status() == Query::Status::INCOMPLETE);
 
     auto result_el = query.result_buffer_elements_nullable();
@@ -876,15 +723,7 @@ TEST_CASE_METHOD(
     // Reset buffers, serialize and resubmit
     set_buffers(query);
     // Submit query
-    rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
 
     REQUIRE(query.query_status() == Query::Status::INCOMPLETE);
     result_el = query.result_buffer_elements_nullable();
@@ -897,15 +736,7 @@ TEST_CASE_METHOD(
     // Reset buffers, serialize and resubmit
     set_buffers(query);
     // Submit query
-    rc = submit_query_wrapper(
-        ctx,
-        array_uri,
-        &query,
-        server_buffers_,
-        serialize_,
-        refactored_query_v2_,
-        finalize_);
-    REQUIRE(rc == TILEDB_OK);
+    query.submit();
 
     REQUIRE(query.query_status() == Query::Status::COMPLETE);
     result_el = query.result_buffer_elements_nullable();
