@@ -118,12 +118,13 @@ VFS::VFS(
 /*                API                */
 /* ********************************* */
 
-std::string VFS::abs_path(const std::string& path) {
+std::string VFS::abs_path(std::string_view path) {
   // workaround for older clang (llvm 3.5) compilers (issue #828)
-  std::string path_copy = path;
+  std::string path_copy(path);
 #ifdef _WIN32
   {
-    std::string norm_sep_path = path_win::slashes_to_backslashes(path);
+    std::string norm_sep_path =
+        path_win::slashes_to_backslashes(std::string(path));
     if (path_win::is_win_path(norm_sep_path))
       return path_win::uri_from_path(Win::abs_path(norm_sep_path));
     else if (URI::is_file(path))
@@ -164,7 +165,8 @@ Status VFS::create_dir(const URI& uri) const {
 #ifdef _WIN32
     return win_.create_dir(uri.to_path());
 #else
-    return posix_.create_dir(uri);
+    posix_.create_dir(uri);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -240,7 +242,8 @@ Status VFS::touch(const URI& uri) const {
 #ifdef _WIN32
     return win_.touch(uri.to_path());
 #else
-    return posix_.touch(uri);
+    posix_.touch(uri);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -252,7 +255,8 @@ Status VFS::touch(const URI& uri) const {
   }
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().touch(uri);
+    s3().touch(uri);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
@@ -285,7 +289,8 @@ Status VFS::cancel_all_tasks() {
 Status VFS::create_bucket(const URI& uri) const {
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().create_bucket(uri);
+    s3().create_bucket(uri);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
@@ -310,7 +315,8 @@ Status VFS::create_bucket(const URI& uri) const {
 Status VFS::remove_bucket(const URI& uri) const {
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().remove_bucket(uri);
+    s3().remove_bucket(uri);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
@@ -335,7 +341,8 @@ Status VFS::remove_bucket(const URI& uri) const {
 Status VFS::empty_bucket(const URI& uri) const {
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().empty_bucket(uri);
+    s3().empty_bucket(uri);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
@@ -361,7 +368,8 @@ Status VFS::is_empty_bucket(
     const URI& uri, [[maybe_unused]] bool* is_empty) const {
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().is_empty_bucket(uri, is_empty);
+    *is_empty = s3().is_empty_bucket(uri);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
@@ -388,7 +396,7 @@ Status VFS::remove_dir(const URI& uri) const {
 #ifdef _WIN32
     return win_.remove_dir(uri.to_path());
 #else
-    return posix_.remove_dir(uri);
+    posix_.remove_dir(uri);
 #endif
   } else if (uri.is_hdfs()) {
 #ifdef HAVE_HDFS
@@ -398,7 +406,7 @@ Status VFS::remove_dir(const URI& uri) const {
 #endif
   } else if (uri.is_s3()) {
 #ifdef HAVE_S3
-    return s3().remove_dir(uri);
+    s3().remove_dir(uri);
 #else
     throw BuiltWithout("S3");
 #endif
@@ -419,6 +427,7 @@ Status VFS::remove_dir(const URI& uri) const {
   } else {
     throw UnsupportedURI(uri.to_string());
   }
+  return Status::Ok();
 }
 
 void VFS::remove_dirs(
@@ -438,7 +447,8 @@ Status VFS::remove_file(const URI& uri) const {
 #ifdef _WIN32
     return win_.remove_file(uri.to_path());
 #else
-    return posix_.remove_file(uri);
+    posix_.remove_file(uri);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -523,7 +533,8 @@ Status VFS::file_size(const URI& uri, uint64_t* size) const {
 #ifdef _WIN32
     return win_.file_size(uri.to_path(), size);
 #else
-    return posix_.file_size(uri, size);
+    posix_.file_size(uri, size);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -661,7 +672,7 @@ Status VFS::is_file(const URI& uri, bool* is_file) const {
 Status VFS::is_bucket(const URI& uri, bool* is_bucket) const {
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    RETURN_NOT_OK(s3().is_bucket(uri, is_bucket));
+    *is_bucket = s3().is_bucket(uri);
     return Status::Ok();
 #else
     *is_bucket = false;
@@ -791,7 +802,8 @@ Status VFS::move_file(const URI& old_uri, const URI& new_uri) {
 #ifdef _WIN32
       return win_.move_path(old_uri.to_path(), new_uri.to_path());
 #else
-      return posix_.move_file(old_uri, new_uri);
+      posix_.move_file(old_uri, new_uri);
+      return Status::Ok();
 #endif
     }
     throw UnsupportedOperation("Moving files");
@@ -860,7 +872,8 @@ Status VFS::move_dir(const URI& old_uri, const URI& new_uri) {
 #ifdef _WIN32
       return win_.move_path(old_uri.to_path(), new_uri.to_path());
 #else
-      return posix_.move_file(old_uri, new_uri);
+      posix_.move_dir(old_uri, new_uri);
+      return Status::Ok();
 #endif
     }
     throw UnsupportedOperation("Moving directories");
@@ -879,12 +892,14 @@ Status VFS::move_dir(const URI& old_uri, const URI& new_uri) {
 
   // S3
   if (old_uri.is_s3()) {
-    if (new_uri.is_s3())
+    if (new_uri.is_s3()) {
 #ifdef HAVE_S3
-      return s3().move_dir(old_uri, new_uri);
+      s3().move_dir(old_uri, new_uri);
+      return Status::Ok();
 #else
       throw BuiltWithout("S3");
 #endif
+    }
     throw UnsupportedOperation("Moving directories");
   }
 
@@ -935,7 +950,8 @@ Status VFS::copy_file(const URI& old_uri, const URI& new_uri) {
 #ifdef _WIN32
       throw VFSException("Copying files on Windows is not yet supported.");
 #else
-      return posix_.copy_file(old_uri, new_uri);
+      posix_.copy_file(old_uri, new_uri);
+      return Status::Ok();
 #endif
     }
     throw UnsupportedOperation("Copying files");
@@ -955,12 +971,14 @@ Status VFS::copy_file(const URI& old_uri, const URI& new_uri) {
 
   // S3
   if (old_uri.is_s3()) {
-    if (new_uri.is_s3())
+    if (new_uri.is_s3()) {
 #ifdef HAVE_S3
-      return s3().copy_file(old_uri, new_uri);
+      s3().copy_file(old_uri, new_uri);
+      return Status::Ok();
 #else
       throw BuiltWithout("S3");
 #endif
+    }
     throw UnsupportedOperation("Copying files");
   }
 
@@ -1000,7 +1018,8 @@ Status VFS::copy_dir(const URI& old_uri, const URI& new_uri) {
       throw VFSException(
           "Copying directories on Windows is not yet supported.");
 #else
-      return posix_.copy_dir(old_uri, new_uri);
+      posix_.copy_dir(old_uri, new_uri);
+      return Status::Ok();
 #endif
     }
     throw UnsupportedOperation("Copying directories");
@@ -1020,12 +1039,14 @@ Status VFS::copy_dir(const URI& old_uri, const URI& new_uri) {
 
   // S3
   if (old_uri.is_s3()) {
-    if (new_uri.is_s3())
+    if (new_uri.is_s3()) {
 #ifdef HAVE_S3
-      return s3().copy_dir(old_uri, new_uri);
+      s3().copy_dir(old_uri, new_uri);
+      return Status::Ok();
 #else
       throw BuiltWithout("S3");
 #endif
+    }
     throw UnsupportedOperation("Copying directories");
   }
 
@@ -1135,7 +1156,8 @@ Status VFS::read_impl(
 #ifdef _WIN32
     return win_.read(uri.to_path(), offset, buffer, nbytes);
 #else
-    return posix_.read(uri, offset, buffer, nbytes, false);
+    posix_.read(uri, offset, buffer, nbytes, false);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -1148,7 +1170,7 @@ Status VFS::read_impl(
   if (uri.is_s3()) {
 #ifdef HAVE_S3
     const auto read_fn = std::bind(
-        &S3::read,
+        &S3::read_impl,
         &s3(),
         std::placeholders::_1,
         std::placeholders::_2,
@@ -1288,7 +1310,8 @@ Status VFS::sync(const URI& uri) {
 #ifdef _WIN32
     return win_.sync(uri.to_path());
 #else
-    return posix_.sync(uri);
+    posix_.sync(uri);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -1378,7 +1401,8 @@ Status VFS::close_file(const URI& uri) {
 #ifdef _WIN32
     return win_.sync(uri.to_path());
 #else
-    return posix_.sync(uri);
+    posix_.sync(uri);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -1439,7 +1463,8 @@ Status VFS::write(
 #ifdef _WIN32
     return win_.write(uri.to_path(), buffer, buffer_size);
 #else
-    return posix_.write(uri, buffer, buffer_size);
+    posix_.write(uri, buffer, buffer_size);
+    return Status::Ok();
 #endif
   }
   if (uri.is_hdfs()) {
@@ -1451,11 +1476,8 @@ Status VFS::write(
   }
   if (uri.is_s3()) {
 #ifdef HAVE_S3
-    if (remote_global_order_write) {
-      s3().global_order_write_buffered(uri, buffer, buffer_size);
-      return Status::Ok();
-    }
-    return s3().write(uri, buffer, buffer_size);
+    s3().write(uri, buffer, buffer_size, remote_global_order_write);
+    return Status::Ok();
 #else
     throw BuiltWithout("S3");
 #endif
