@@ -353,17 +353,19 @@ Status Reader::apply_query_condition(
   if (stride == UINT64_MAX)
     stride = 1;
 
+  QueryCondition::Params params(query_memory_tracker_, array_schema_);
   if (condition_.has_value()) {
     RETURN_NOT_OK(condition_->apply(
-        array_schema_, fragment_metadata_, result_cell_slabs, stride));
+        params, fragment_metadata_, result_cell_slabs, stride));
   }
 
   // Apply delete conditions.
   if (!delete_and_update_conditions_.empty()) {
     for (uint64_t i = 0; i < delete_and_update_conditions_.size(); i++) {
       // For legacy, always run the timestamped condition.
+
       RETURN_NOT_OK(timestamped_delete_and_update_conditions_[i].apply(
-          array_schema_, fragment_metadata_, result_cell_slabs, stride));
+          params, fragment_metadata_, result_cell_slabs, stride));
     }
   }
 
@@ -371,7 +373,7 @@ Status Reader::apply_query_condition(
   if (!delete_timestamps_condition_.empty()) {
     // Remove cells with partial overlap from the bitmap.
     RETURN_NOT_OK(delete_timestamps_condition_.apply(
-        array_schema_, fragment_metadata_, result_cell_slabs, stride));
+        params, fragment_metadata_, result_cell_slabs, stride));
   }
 
   return Status::Ok();
@@ -453,7 +455,9 @@ Status Reader::compute_range_result_coords(
         result_coords.emplace_back(tile, pos);
     }
   } else {  // Sparse
-    std::vector<uint8_t> result_bitmap(coords_num, 1);
+    auto resource =
+        query_memory_tracker_->get_resource(MemoryType::RESULT_TILE_BITMAP);
+    tdb::pmr::vector<uint8_t> result_bitmap(coords_num, 1, resource);
 
     // Compute result and overwritten bitmap per dimension
     for (unsigned d = 0; d < dim_num; ++d) {
@@ -474,8 +478,10 @@ Status Reader::compute_range_result_coords(
     // Apply partial overlap condition, if required.
     const auto frag_meta = fragment_metadata_[tile->frag_idx()];
     if (process_partial_timestamps(*frag_meta)) {
+      QueryCondition::Params params(
+          query_memory_tracker_, *(frag_meta->array_schema().get()));
       RETURN_NOT_OK(partial_overlap_condition_.apply_sparse<uint8_t>(
-          *(frag_meta->array_schema().get()), *tile, result_bitmap));
+          params, *tile, result_bitmap));
     }
 
     // Gather results
@@ -492,7 +498,7 @@ Status Reader::compute_range_result_coords(
     Subarray& subarray,
     const std::vector<bool>& single_fragment,
     const std::map<std::pair<unsigned, uint64_t>, size_t>& result_tile_map,
-    std::vector<ResultTile>& result_tiles,
+    IndexedList<ResultTile>& result_tiles,
     std::vector<std::vector<ResultCoords>>& range_result_coords) {
   auto timer_se = stats_->start_timer("compute_range_result_coords");
 
@@ -547,7 +553,7 @@ Status Reader::compute_range_result_coords(
     uint64_t range_idx,
     uint32_t fragment_idx,
     const std::map<std::pair<unsigned, uint64_t>, size_t>& result_tile_map,
-    std::vector<ResultTile>& result_tiles,
+    IndexedList<ResultTile>& result_tiles,
     std::vector<ResultCoords>& range_result_coords) {
   // Skip dense fragments
   if (fragment_metadata_[fragment_idx]->dense())
@@ -568,12 +574,14 @@ Status Reader::compute_range_result_coords(
         auto tile_it = result_tile_map.find(pair);
         assert(tile_it != result_tile_map.end());
         auto tile_idx = tile_it->second;
-        auto& tile = result_tiles[tile_idx];
+
+        auto tile = result_tiles.begin();
+        std::advance(tile, tile_idx);
 
         // Add results only if the sparse tile MBR is not fully
         // covered by a more recent fragment's non-empty domain
         if (!sparse_tile_overwritten(fragment_idx, i))
-          RETURN_NOT_OK(get_all_result_coords(&tile, range_result_coords));
+          RETURN_NOT_OK(get_all_result_coords(&*tile, range_result_coords));
       }
       ++tr;
     } else {
@@ -582,15 +590,16 @@ Status Reader::compute_range_result_coords(
       auto tile_it = result_tile_map.find(pair);
       assert(tile_it != result_tile_map.end());
       auto tile_idx = tile_it->second;
-      auto& tile = result_tiles[tile_idx];
+      auto tile = result_tiles.begin();
+      std::advance(tile, tile_idx);
       if (t->second == 1.0) {  // Full overlap
         // Add results only if the sparse tile MBR is not fully
         // covered by a more recent fragment's non-empty domain
         if (!sparse_tile_overwritten(fragment_idx, t->first))
-          RETURN_NOT_OK(get_all_result_coords(&tile, range_result_coords));
+          RETURN_NOT_OK(get_all_result_coords(&*tile, range_result_coords));
       } else {  // Partial overlap
         RETURN_NOT_OK(compute_range_result_coords(
-            subarray, fragment_idx, &tile, range_idx, range_result_coords));
+            subarray, fragment_idx, &*tile, range_idx, range_result_coords));
       }
       ++t;
     }
@@ -603,7 +612,7 @@ Status Reader::compute_range_result_coords(
     Subarray& subarray,
     uint64_t range_idx,
     const std::map<std::pair<unsigned, uint64_t>, size_t>& result_tile_map,
-    std::vector<ResultTile>& result_tiles,
+    IndexedList<ResultTile>& result_tiles,
     std::vector<ResultCoords>& range_result_coords) {
   // Gather result range coordinates per fragment
   auto fragment_num = fragment_metadata_.size();
@@ -678,7 +687,7 @@ Status Reader::compute_subarray_coords(
 }
 
 Status Reader::compute_sparse_result_tiles(
-    std::vector<ResultTile>& result_tiles,
+    IndexedList<ResultTile>& result_tiles,
     std::map<std::pair<unsigned, uint64_t>, size_t>* result_tile_map,
     std::vector<bool>* single_fragment) {
   auto timer_se = stats_->start_timer("compute_sparse_result_tiles");
@@ -711,7 +720,8 @@ Status Reader::compute_sparse_result_tiles(
           auto pair = std::pair<unsigned, uint64_t>(f, t);
           // Add tile only if it does not already exist
           if (result_tile_map->find(pair) == result_tile_map->end()) {
-            result_tiles.emplace_back(f, t, *fragment_metadata_[f].get());
+            result_tiles.emplace_back(
+                f, t, *fragment_metadata_[f].get(), query_memory_tracker_);
             (*result_tile_map)[pair] = result_tiles.size() - 1;
           }
           // Always check range for multiple fragments or fragments with
@@ -730,7 +740,8 @@ Status Reader::compute_sparse_result_tiles(
         auto pair = std::pair<unsigned, uint64_t>(f, t);
         // Add tile only if it does not already exist
         if (result_tile_map->find(pair) == result_tile_map->end()) {
-          result_tiles.emplace_back(f, t, *fragment_metadata_[f].get());
+          result_tiles.emplace_back(
+              f, t, *fragment_metadata_[f].get(), query_memory_tracker_);
           (*result_tile_map)[pair] = result_tiles.size() - 1;
         }
         // Always check range for multiple fragments or fragments with
@@ -1564,7 +1575,7 @@ Status Reader::compute_result_cell_slabs_global(
     tile_subarrays.emplace_back(
         subarray.crop_to_tile((const T*)&tc[0], cell_order));
     auto& tile_subarray = tile_subarrays.back();
-    throw_if_not_ok(tile_subarray.template compute_tile_coords<T>());
+    tile_subarray.template compute_tile_coords<T>();
 
     RETURN_NOT_OK(compute_result_cell_slabs_row_col<T>(
         tile_subarray,
@@ -1580,7 +1591,7 @@ Status Reader::compute_result_cell_slabs_global(
 }
 
 Status Reader::compute_result_coords(
-    std::vector<ResultTile>& result_tiles,
+    IndexedList<ResultTile>& result_tiles,
     std::vector<ResultCoords>& result_coords) {
   auto timer_se = stats_->start_timer("compute_result_coords");
 
@@ -1719,7 +1730,7 @@ Status Reader::dense_read() {
   // `sparse_result_tiles` will hold all the relevant result tiles of
   // sparse fragments
   std::vector<ResultCoords> result_coords;
-  std::vector<ResultTile> sparse_result_tiles;
+  IndexedList<ResultTile> sparse_result_tiles(query_memory_tracker_);
   RETURN_NOT_OK(compute_result_coords(sparse_result_tiles, result_coords));
 
   // Compute result cell slabs.
@@ -1731,7 +1742,7 @@ Status Reader::dense_read() {
   std::vector<ResultTile*> result_tiles;
   auto& subarray = read_state_.partitioner_.current();
 
-  RETURN_NOT_OK(subarray.compute_tile_coords<T>());
+  subarray.compute_tile_coords<T>();
   RETURN_NOT_OK(compute_result_cell_slabs<T>(
       subarray,
       result_space_tiles,
@@ -1780,9 +1791,12 @@ Status Reader::get_all_result_coords(
       array_->timestamp_start(), array_->timestamp_end_opened_at());
   if (fragment_metadata_[tile->frag_idx()]->has_timestamps() &&
       partial_overlap) {
-    std::vector<uint8_t> result_bitmap(coords_num, 1);
+    auto resource =
+        query_memory_tracker_->get_resource(MemoryType::RESULT_TILE_BITMAP);
+    tdb::pmr::vector<uint8_t> result_bitmap(coords_num, 1, resource);
+    QueryCondition::Params params(query_memory_tracker_, array_schema_);
     RETURN_NOT_OK(partial_overlap_condition_.apply_sparse<uint8_t>(
-        *(frag_meta->array_schema().get()), *tile, result_bitmap));
+        params, *tile, result_bitmap));
 
     for (uint64_t i = 0; i < coords_num; ++i) {
       if (result_bitmap[i]) {
@@ -1975,7 +1989,7 @@ Status Reader::sparse_read() {
   // `sparse_result_tiles` will hold all the relevant result tiles of
   // sparse fragments
   std::vector<ResultCoords> result_coords;
-  std::vector<ResultTile> sparse_result_tiles;
+  IndexedList<ResultTile> sparse_result_tiles(query_memory_tracker_);
 
   RETURN_NOT_OK(compute_result_coords(sparse_result_tiles, result_coords));
   std::vector<ResultTile*> result_tiles;
@@ -2057,7 +2071,7 @@ bool Reader::sparse_tile_overwritten(
   return false;
 }
 
-void Reader::erase_coord_tiles(std::vector<ResultTile>& result_tiles) const {
+void Reader::erase_coord_tiles(IndexedList<ResultTile>& result_tiles) const {
   for (auto& tile : result_tiles) {
     auto dim_num = array_schema_.dim_num();
     for (unsigned d = 0; d < dim_num; ++d)
