@@ -70,10 +70,10 @@ SparseIndexReaderBase::SparseIndexReaderBase(
     StrategyParams& params,
     bool include_coords)
     : ReaderBase(stats, logger, params)
+    , read_state_(array_->fragment_metadata().size())
     , tmp_read_state_(array_->fragment_metadata().size())
     , memory_budget_(config_, reader_string)
     , include_coords_(include_coords)
-    , array_memory_tracker_(params.memory_tracker())
     , memory_used_for_coords_total_(0)
     , deletes_consolidation_no_purge_(
           buffers_.count(constants::delete_timestamps) != 0)
@@ -131,13 +131,13 @@ SparseIndexReaderBase::SparseIndexReaderBase(
 /*        PROTECTED METHODS       */
 /* ****************************** */
 
-const typename SparseIndexReaderBase::ReadState*
+const typename SparseIndexReaderBase::ReadState&
 SparseIndexReaderBase::read_state() const {
-  return &read_state_;
+  return read_state_;
 }
 
-typename SparseIndexReaderBase::ReadState* SparseIndexReaderBase::read_state() {
-  return &read_state_;
+void SparseIndexReaderBase::set_read_state(ReadState read_state) {
+  read_state_ = std::move(read_state);
 }
 
 uint64_t SparseIndexReaderBase::available_memory() {
@@ -319,11 +319,10 @@ Status SparseIndexReaderBase::load_initial_data() {
   }
 
   auto timer_se = stats_->start_timer("load_initial_data");
-  read_state_.done_adding_result_tiles_ = false;
+  read_state_.set_done_adding_result_tiles(false);
 
   // For easy reference.
   const auto dim_num = array_schema_.dim_num();
-  auto fragment_num = fragment_metadata_.size();
 
   // Load delete conditions.
   auto&& [st, conditions, update_values] =
@@ -366,9 +365,6 @@ Status SparseIndexReaderBase::load_initial_data() {
     }
   }
 
-  // Make sure there is enough space for tiles data.
-  read_state_.frag_idx_.resize(fragment_num);
-
   // Calculate ranges of tiles in the subarray, if set.
   if (subarray_.is_set()) {
     // At this point, full memory budget is available.
@@ -383,15 +379,22 @@ Status SparseIndexReaderBase::load_initial_data() {
     // Tile ranges computation will not stop if it exceeds memory budget.
     // This is ok as it is a soft limit and will be taken into consideration
     // below.
-    RETURN_NOT_OK(subarray_.precompute_all_ranges_tile_overlap(
+    subarray_.precompute_all_ranges_tile_overlap(
         storage_manager_->compute_tp(),
-        read_state_.frag_idx_,
-        &tmp_read_state_));
+        read_state_.frag_idx(),
+        &tmp_read_state_);
 
     if (tmp_read_state_.memory_used_tile_ranges() >
         memory_budget_.ratio_tile_ranges() * memory_budget_.total_budget())
       return logger_->status(
           Status_ReaderError("Exceeded memory budget for result tile ranges"));
+  } else {
+    for (const auto& [name, _] : aggregates_) {
+      if (array_schema_.is_dim(name)) {
+        subarray_.load_relevant_fragment_rtrees(storage_manager_->compute_tp());
+        break;
+      }
+    }
   }
 
   // Compute tile offsets to load and var size to load for attributes.
@@ -490,7 +493,13 @@ void SparseIndexReaderBase::load_tile_offsets_for_fragments(
   load_tile_offsets(relevant_fragments, attr_tile_offsets_to_load_);
 
   // Load tile metadata.
-  load_tile_metadata(relevant_fragments, attr_tile_offsets_to_load_);
+  auto md_names_to_load = attr_tile_offsets_to_load_;
+  for (const auto& [name, _] : aggregates_) {
+    if (array_schema_.is_dim(name)) {
+      md_names_to_load.emplace_back(name);
+    }
+  }
+  load_tile_metadata(relevant_fragments, md_names_to_load);
 }
 
 Status SparseIndexReaderBase::read_and_unfilter_coords(
@@ -622,7 +631,8 @@ void SparseIndexReaderBase::compute_tile_bitmaps(
           auto& ranges_for_dim = subarray_.ranges_for_dim(dim_idx);
 
           // Compute the list of range index to process.
-          std::vector<uint64_t> relevant_ranges;
+          tdb::pmr::vector<uint64_t> relevant_ranges(
+              query_memory_tracker_->get_resource(MemoryType::DIMENSIONS));
           relevant_ranges.reserve(ranges_for_dim.size());
           domain.dimension_ptr(dim_idx)->relevant_ranges(
               ranges_for_dim, mbr[dim_idx], relevant_ranges);
@@ -631,9 +641,8 @@ void SparseIndexReaderBase::compute_tile_bitmaps(
           // there is no need to compute bitmaps.
           const bool non_overlapping = std::is_same<BitmapType, uint8_t>::value;
           if (non_overlapping) {
-            std::vector<bool> covered_bitmap =
-                domain.dimension_ptr(dim_idx)->covered_vec(
-                    ranges_for_dim, mbr[dim_idx], relevant_ranges);
+            auto covered_bitmap = domain.dimension_ptr(dim_idx)->covered_vec(
+                ranges_for_dim, mbr[dim_idx], relevant_ranges);
 
             // See if any range is covered.
             uint64_t count = std::accumulate(
@@ -718,8 +727,10 @@ void SparseIndexReaderBase::apply_query_condition(
             }
 
             // Remove cells with partial overlap from the bitmap.
+            QueryCondition::Params params(
+                query_memory_tracker_, *(frag_meta->array_schema().get()));
             RETURN_NOT_OK(partial_overlap_condition_.apply_sparse<BitmapType>(
-                *(frag_meta->array_schema().get()), *rt, rt->bitmap()));
+                params, *rt, rt->bitmap()));
             rt->count_cells();
           }
 
@@ -733,10 +744,10 @@ void SparseIndexReaderBase::apply_query_condition(
           if (frag_meta->has_delete_meta() &&
               !deletes_consolidation_no_purge_) {
             // Remove cells deleted cells using the open timestamp.
+            QueryCondition::Params params(
+                query_memory_tracker_, *(frag_meta->array_schema().get()));
             RETURN_NOT_OK(delete_timestamps_condition_.apply_sparse<BitmapType>(
-                *(frag_meta->array_schema().get()),
-                *rt,
-                rt->post_dedup_bitmap()));
+                params, *rt, rt->post_dedup_bitmap()));
             if (array_schema_.allows_dups()) {
               rt->count_cells();
             }
@@ -744,10 +755,10 @@ void SparseIndexReaderBase::apply_query_condition(
 
           // Compute the result of the query condition for this tile.
           if (condition_.has_value()) {
+            QueryCondition::Params params(
+                query_memory_tracker_, *(frag_meta->array_schema().get()));
             RETURN_NOT_OK(condition_->apply_sparse<BitmapType>(
-                *(frag_meta->array_schema().get()),
-                *rt,
-                rt->post_dedup_bitmap()));
+                params, *rt, rt->post_dedup_bitmap()));
             if (array_schema_.allows_dups()) {
               rt->count_cells();
             }
@@ -775,19 +786,20 @@ void SparseIndexReaderBase::apply_query_condition(
                 // start.
                 if (delete_timestamp >= frag_meta->timestamp_range().first) {
                   // Apply timestamped condition or regular condition.
+                  QueryCondition::Params params(
+                      query_memory_tracker_,
+                      *(frag_meta->array_schema().get()));
                   if (!frag_meta->has_timestamps() ||
                       delete_timestamp > frag_meta->timestamp_range().second) {
-                    RETURN_NOT_OK(delete_and_update_conditions_[i]
-                                      .apply_sparse<BitmapType>(
-                                          *(frag_meta->array_schema().get()),
-                                          *rt,
-                                          rt->post_dedup_bitmap()));
+                    RETURN_NOT_OK(
+                        delete_and_update_conditions_[i]
+                            .apply_sparse<BitmapType>(
+                                params, *rt, rt->post_dedup_bitmap()));
                   } else {
-                    RETURN_NOT_OK(timestamped_delete_and_update_conditions_[i]
-                                      .apply_sparse<BitmapType>(
-                                          *(frag_meta->array_schema().get()),
-                                          *rt,
-                                          rt->post_dedup_bitmap()));
+                    RETURN_NOT_OK(
+                        timestamped_delete_and_update_conditions_[i]
+                            .apply_sparse<BitmapType>(
+                                params, *rt, rt->post_dedup_bitmap()));
                   }
 
                   if (deletes_consolidation_no_purge_) {
